--[[

   JPEGImage class

   This class provides public functions required to determine the postion of the
   EXIF tag UserComment within a JPEG image and for writing a new text to the
   tag.

   To understand the code it is recommended to study the following documents:

         TsuruZoh Tachibanaya, "Description of Exif file format"
         http://park2.wakwak.com/~tsuruzoh/Computer/Digicams/exif-e.html

         Ger Vermeulen, "Description of Exif file format"
         http://www.gvsoft.homedns.org/exif/Exif-explanation.html

         JEITA CP-3451, Exchangeable image file format for digital still cameras: Exif Version 2.2
         http://www.exif.org/Exif2-2.PDF

         TIFF, Revision 6.0, Final  June 3, 1992
         http://www.exif.org/TIFF6.pdf


    Author: Stefan Dittforth, (stift at email.de)

    Change History:

        v0.0.1 (15/11/2009) - First version, code moved from TagMe.lua to EXIFModule.lua for
                              reuse in other scripts.

]]

JPEGImage = {}

------------------------------------------------------------------------------------
--
-- JPEGImage:new() function
--
-- Contains the code to crawl throught the JPEG image file and find the data block
-- with EXIF tag information.
--
-- Input:
--
--      imageFilePath = full path to image file
--
-- Output:
--
--      self.UserComment     = current content of UserComment EXIF tag
--      self.DataStartInFile = start position of EXIF tag in image file
--      self.NumOfComponents = maximum number of characters allowed for EXIF tag
--
-------------------------------------------------------------------------------------

function JPEGImage:new(imageFilePath)

    --
    -- private attributes
    --
    self.FilePath = imageFilePath

    local EXIFDataFormatTable = {}
    EXIFDataFormatTable[1] = { [1] = 1, [2] = "unsigned byte" }
    EXIFDataFormatTable[2] = { [1] = 1, [2] = "ascii strings" }
    EXIFDataFormatTable[3] = { [1] = 2, [2] = "unsigned short" }
    EXIFDataFormatTable[4] = { [1] = 4, [2] = "unsigned long" }
    EXIFDataFormatTable[5] = { [1] = 8, [2] = "unsigned rational" }
    EXIFDataFormatTable[6] = { [1] = 1, [2] = "signed byte" }
    EXIFDataFormatTable[7] = { [1] = 1, [2] = "undefined" }
    EXIFDataFormatTable[8] = { [1] = 2, [2] = "signed short" }
    EXIFDataFormatTable[9] = { [1] = 4, [2] = "signed long" }
    EXIFDataFormatTable[10] = { [1] = 8, [2] = "signed rational" }
    EXIFDataFormatTable[11] = { [1] = 4, [2] = "single float" }
    EXIFDataFormatTable[12] = { [1] = 8, [2] = "double float" }

    --
    -- private methodes
    --

    local function ConvertToBigEndian(InputString)

        local ReturnString = ""
        local InputLength = string.len(InputString)

        for i = 1, InputLength do
            ReturnString = ReturnString..InputString:sub(InputLength-(i-1), InputLength-(i-1))
        end

        return ReturnString

    end

    local function StringToNumber(InputString)

        local ReturnNumber, InputLength, i

        ReturnNumber = 0
        InputLength = string.len(InputString)

        for i = 1, InputLength do
            ReturnNumber = ReturnNumber + InputString:byte(i)*2^((InputLength-i)*8)
        end

        return ReturnNumber

    end

    local function StringToHexCodes(InputString)

        local i, ReturnString

        ReturnString = ""
        for i = 1, string.len(InputString) do
            ReturnString = ReturnString..string.format("%02X", InputString:byte(i))
        end

        return ReturnString

    end

    local function GetBytes(InputString, BigEndian)

        local ReturnString = ""

        if not BigEndian then
            ReturnString = ConvertToBigEndian(InputString)
        else
            ReturnString = InputString
        end

        return ReturnString

    end

    -------------------------------------------------------------------------------------
    --
    -- readIFD (TIFFData, StartIFD)
    --
    -- Input:
    --        TIFFData - TIFF data block, contains all EXIF tags
    --        StartIFD - startposition of IFD in TIFF data block, there can be more than
    --                   one IFD in a TIFF data block
    --
    -- Output: Offset to next IFD
    --
    -- Purpose:
    --         Extracts the EXIF tags from an individual IFD
    --
    -------------------------------------------------------------------------------------

    local function readIFD (TIFFData, StartIFD)

        local NumOfIFDEntries, LenghtIFDEntry, Offset, f, StartNextIFD
        local TagID, DataFormat, NumOfComponents, DataValue, DataStartInFile, FormatDescription

        -- first two bytes of IFD contains number of directory entries
        NumOfIFDEntries = StringToNumber(GetBytes(TIFFData:sub(StartIFD, StartIFD + 1), BigEndian))

        -- read all IFD directory entries (12 bytes per entry)
        LenghtIFDEntry = 12
        Offset = StartIFD + 2

        for i = 0, NumOfIFDEntries - 1 do

            -- get tag number, format and number of components
            TagID               = StringToHexCodes(GetBytes(TIFFData:sub(Offset, Offset + 1), BigEndian))
            DataFormat        = StringToNumber(GetBytes(TIFFData:sub(Offset + 2, Offset + 3), BigEndian))
            FormatDescription = EXIFDataFormatTable[DataFormat][2]
            NumOfComponents   = StringToNumber(GetBytes(TIFFData:sub(Offset + 4, Offset + 7), BigEndian))

            -- get the EXIF data value
            if NumOfComponents * EXIFDataFormatTable[DataFormat][1] > 4 then
                -- next 4 bytes contain an offset to the data
                OffsetToEXIFDataValue = StringToNumber(GetBytes(TIFFData:sub(Offset+8, Offset+11), BigEndian))
                DataValue = TIFFData:sub(1 + OffsetToEXIFDataValue, 1 + OffsetToEXIFDataValue + NumOfComponents)
                DataStartInFile = self.JPEGImage.TIFFHeaderStart + OffsetToEXIFDataValue
            else
                -- next 4 bytes are the data
                DataValue = GetBytes(TIFFData:sub(Offset + 8, Offset + 11), BigEndian)
                DataStartInFile = self.JPEGImage.TIFFHeaderStart + Offset + 8 - 1 -- "- 1" because offset in file starts with 0, in TIFFData it starts with 1
            end

             -- convert EXIF data value
             f = FormatDescription
             if f == "unsigned byte" or f == "unsigned short" or f == "unsigned long" or f == "unsigned rational" then
                 DataValue = StringToNumber(DataValue)
             end

            -- check if we got the EXIF UserComment tag (9286.H)
            if TagID == "9286" then

                -- extract the UserComment string and the start position in the image file
                self.UserComment = DataValue:sub(9, DataValue:len()) -- according to JPEG specificiation first 8 bytes specify code designation
                self.DataStartInFile = DataStartInFile + 8
                self.NumOfComponents = NumOfComponents - 8 -- less then the 8 code designation bytes

                -- we got what we wanted, finish the loop
                break

            end

            -- check if there is an EXIF IFD tag (pointer to EXIF IFD)
            if TagID == "8769" then
                readIFD(TIFFData, DataValue + 1)
            end

             Offset = Offset + LenghtIFDEntry

        end

        -- retun offset to next IFD
        return StringToNumber(GetBytes(TIFFData:sub(Offset, Offset+3), BigEndian))

    end

    -------------------------------------------------------------------------------------
    --
    -- extractEXIFTags (TIFFData)
    --
    -- Input:
    --        TIFFData - TIFF data block, contains all EXIF tags
    --
    -- Output: none
    --
    -- Purpose:
    --         Steps through the IFDs
    --
    -------------------------------------------------------------------------------------

    local function extractEXIFTags(TIFFData)

        local BigEndian, ByteAlignTagMark, StartNextIFD

        -- read byte align from TIFF header
        -- first 2 bytes of TIFF header defines byte align of TIFF data.
        -- 0x4949='I I' is 'Intel' type byte align, 0x4d4d='MM' is 'Motorola' type byte align.
        if StringToHexCodes(TIFFData:sub(1,2)) == "4D4D" then
            BigEndian = true
        elseif StringToHexCodes(TIFFData:sub(1,2)) == "4949" then
            BigEndian = false
        else
            print("byte align setting of TIFF data bytes not found!")
            return nil
        end

        -- test byte align is correct by reading the tag marker
        ByteAlignTagMark = StringToHexCodes(TIFFData:sub(3,4))
        if (BigEndian and ByteAlignTagMark == "002A") or (not BigEndian and ByteAlignTagMark == "2A00") then

            -- the last 4 bytes of TIFF header is the offset to the first IFD
            -- after TIFF header comes first IFD (Image File Directory) aka IFD0
            StartNextIFD = StringToNumber(GetBytes(TIFFData:sub(5,8), BigEndian))

            -- read all IFDs
            repeat
                StartNextIFD = readIFD(TIFFData, StartNextIFD + 1) -- "+ 1" because offset values start from 0 but data block starts from 1
            until StartNextIFD == 0

        else
            print("byte align tag mark not as expected! tag mark = ", ByteAlignTagMark)
            return nil
        end

    end

    -------------------------------------------------------------------------------------
    --
    -- loadEXIFData ()
    --
    -- Checks whether file is a valid JPEG with EXIF data block. Reads TIFF data block
    -- into memory and calls function to extract EXIF tag data.
    --
    -------------------------------------------------------------------------------------

    local function loadEXIFData()

        local inp, data, JPEGMarker, EXIFDataFound, EXIFBlockSize, EXIFData, TIFFData

        inp = io.open(self.FilePath, "rb")
        data = inp:read(2)
        if not data then
            print("could not open file ", self.filePath)
            inp:close()
            return nil
        else
            JPEGMarker = StringToHexCodes(data:sub(1, 2))
            -- the first two bytes of a JPEG file must be "FFD8", the JPEG image marker
            if JPEGMarker == "FFD8" then
                -- search for the EXIF marker "FFE1"
                EXIFDataFound = false
                repeat
                    data = inp:read(1)
                    if data then JPEGMarker = StringToHexCodes(data:sub(1,1)) end
                    if JPEGMarker == "FF" then -- start of a JPEG marker
                        data = inp:read(1)
                        JPEGMarker = StringToHexCodes(data:sub(1,1))
                        if JPEGMarker == "E1" then -- start of EXIF marker found
                            -- initialise the data structure
                            self.JPEGImage = {}
                            -- get the size of the EXIF data block
                            data = inp:read(2)
                            EXIFBlockSize = StringToNumber(data:sub(1, 2))
                            -- check for valid start of EXIF block
                            data = inp:read(6)
                            if StringToHexCodes((data:sub(1, 6))) == "457869660000" then -- valid EXIF block found
                                self.JPEGImage.TIFFHeaderStart = inp:seek() -- save current file position, this is the start of the TIFF header
                                TIFFData = inp:read(EXIFBlockSize - 2 - 6)
                                EXIFDataFound = true
                                -- now that we have loaded all EXIF data close the file
                                inp:close()
                                -- extract EXIF data
                                extractEXIFTags(TIFFData)
                            else
                                print("no valid start of EXIF block found in "..self.filePath.."!")
                                return nil
                            end
                        end
                    end
                until EXIFDataFound or not data
            else
                print("file "..self.filePath.." is not a JPEG image")
                inp:close()
                return nil
            end
        end

    end -- loadEXIFData = function ()

    -- load EXIF data into data structure
    loadEXIFData()

end

------------------------------------------------------------------------------------
--
-- JPEGImage:SetUserComment() function
--
-- Sets a new user comment
--
-- Input:
--
--      NewUserComment = new comment string
--
-- Output:
--
--      Modifies image file set with JPEGImage:new() function
--
-------------------------------------------------------------------------------------
function JPEGImage:SetUserComment(NewUserComment)

    local ImageFile, F

    -- Reduce new comment to maximum lenght of UserComment field, if we don't do this we will
    -- corrupt the data structure if the image file.
    -- It looks like on a Canon SX10 IS camera the user comment can be up to 255 characters long.
    if NewUserComment:len() > self.NumOfComponents-1 then NewUserComment = NewUserComment:sub(1, self.NumOfComponents-1) end

    -- Save the time of last access and last modification before writing to the UserComment field
    -- of the image.
    F = os.stat(self.FilePath)

    ImageFile = io.open(self.FilePath, "r+")
    -- move forward to the start position of the user comment field
    ImageFile:seek("set", self.DataStartInFile)
    ImageFile:write(NewUserComment, string.rep(string.char(0), self.NumOfComponents - NewUserComment:len())) -- overwrite the reminder with zeros
    ImageFile:close()

    -- Restore time of last access and last modification.
    -- When writing the new UserComment filed the image file gets a new new "last
    -- mofified" time stamp. With restoring the original time stamps the script
    -- keeps the modifications to an image file to a minimum.
    os.utime(self.FilePath, F["atime"], F["mtime"])

end
